/*
    Copyright 2004-2011 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "blowfish.h"

typedef unsigned char   u8;
typedef unsigned short  u16;
typedef unsigned int    u32;



#define VER         "0.2"
#define BUFFSZ      256
#define VISFMT      "%20"



void readshow(FILE *fd, u8 *msg, int dec);
u8 *cuteftp_pwd(u8 *pwd, u8 size, int algo);
void read_err(void);
void std_err(void);
void myexit(void);



int main(int argc, char *argv[]) {
    FILE    *fd;
    u32     b4;
    int     i;
    u16     b2,
            hosts;
    u8      b;

    setbuf(stdout, NULL);

    fputs("\n"
        "CuteFTP data retriever and password decoder "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stdout);

    if(argc < 2) {
        printf("\n"
            "Usage: %s <SM.DAT/hex_password>\n"
            "\n"
            "sm.dat is located in the folders inside %%APPDATA%%\\GlobalSCAPE\n"
            "\n", argv[0]);
        myexit();
    }

    fd = fopen(argv[1], "rb");
    //if(!fd) std_err();
    if(!fd) {
        readshow(NULL, argv[1], -1);
        myexit();
    }

    // quick raw password
    fseek(fd, 0, SEEK_END);
    i = ftell(fd);
    fseek(fd, 0, SEEK_SET);
    if(i <= 32) {
        readshow(fd, "raw_password:", -1);
        fclose(fd);
        myexit();
    }

    if(fread(&b, 1, 1, fd) != 1) read_err();
    if(fseek(fd, b, SEEK_SET)) std_err();
    readshow(fd, "Main Folder", 0);
    if(fseek(fd, 14, SEEK_CUR)) std_err();
    if(fread(&hosts, 2, 1, fd) != 1) read_err();
    if(fseek(fd, 4 + (hosts << 2), SEEK_CUR)) std_err();
    if(fread(&b2, 2, 1, fd) != 1) read_err();
    if(fseek(fd, b2, SEEK_CUR)) std_err();

    for(;;) {
        printf("--------------------\n");
        for(i = 0; i < 3; i++) {
            if(fread(&b4, 4, 1, fd) != 1) break;
        }
        if(feof(fd)) break;     // lame but this is only an experiment
        if(fseek(fd, 4, SEEK_CUR)) std_err();
        if(b4) {
            readshow(fd, "Main Folder", 0);
            if(fseek(fd, 14, SEEK_CUR)) std_err();
            if(fread(&b2, 2, 1, fd) != 1) read_err();
            if(fseek(fd, b2 << 2, SEEK_CUR)) std_err();
            if(fseek(fd, 2, SEEK_CUR)) std_err();
            continue;
        }

        readshow(fd, "Label", 0);
        if(fseek(fd, 32, SEEK_CUR)) std_err();
        readshow(fd, "Timezone", 0);
        if(fseek(fd, 8, SEEK_CUR)) std_err();
        readshow(fd, "Hostname", 0);
        if(fread(&b, 1, 1, fd) != 1) read_err();
        if(b == 1) {
            readshow(fd, "", 0);
            readshow(fd, "", 0);
            readshow(fd, "", 0);
            readshow(fd, "", 0);
            readshow(fd, "", 0);
            readshow(fd, "Comment", 0);
            if(fseek(fd, 20, SEEK_CUR)) std_err();
        } else {
            readshow(fd, "", 0);
            readshow(fd, "", 0);
            readshow(fd, "", 0);
            readshow(fd, "Username", 1);
            readshow(fd, "Password", 1);
            readshow(fd, "Comment", 0);
            if(fseek(fd, 4, SEEK_CUR)) std_err();
            if(fread(&b2, 2, 1, fd) != 1) read_err();
            printf(VISFMT"s %hu\n", "Port", b2);
            if(fseek(fd, 14, SEEK_CUR)) std_err();
        }
        readshow(fd, "Remote folder", 0);
        readshow(fd, "Local folder", 0);
        if(fseek(fd, 20, SEEK_CUR)) std_err();
        readshow(fd, "remote filter", 0);
        if(fread(&b2, 2, 1, fd) != 1) read_err();
        for(i = 0; i < b2; i++) {
            readshow(fd, "filter mask", 0);
        }
        if(fread(&b2, 2, 1, fd) != 1) read_err();
        for(i = 0; i < b2; i++) {
            readshow(fd, "filter not mask", 0);
        }
        if(fseek(fd, 14, SEEK_CUR)) std_err();
        readshow(fd, "Proxy host", 0);
        if(fseek(fd, 4, SEEK_CUR)) std_err();
        readshow(fd, "Proxy user", 0);
        readshow(fd, "Proxy password", 0);
        if(fseek(fd, 14, SEEK_CUR)) std_err();
        if(fread(&b2, 2, 1, fd) != 1) read_err();
        for(i = 0; i < b2; i++) {
            readshow(fd, "Proxy advanced", 0);
        }
        if(fseek(fd, 2, SEEK_CUR)) std_err();
        readshow(fd, "Socks host", 0);
        if(fseek(fd, 4, SEEK_CUR)) std_err();
        readshow(fd, "Socks user", 0);
        readshow(fd, "Socks password", 0);
        if(fseek(fd, 22, SEEK_CUR)) std_err();
    }

    fclose(fd);
    myexit();
    return(0);
}



void readshow(FILE *fd, u8 *msg, int dec) {
    int     t,
            len;
    u8      buff[BUFFSZ + 1],
            b,
            *p;

    if(dec < 0) {
        if(!fd) {
            for(len = 0; len < BUFFSZ; len++) {
                if(sscanf(msg + (len * 2), "%02x", &t) != 1) break;
                buff[len] = t;
            }
            msg = "hex_password:";
        } else {
            len = fread(buff, 1, BUFFSZ, fd);
        }
    } else {
        if(fread(&b, 1, 1, fd) != 1) read_err();
        len = b;
        if(b == 0xff) {
            if(fread(&b, 1, 1, fd) != 1) read_err();
            len = b;
            if(fread(&b, 1, 1, fd) != 1) read_err();
            len |= (b << 8);
        }
        if(len <= 0) return;

        if(len > BUFFSZ) {  // blah, I don't want to waste 64kb of memory
            if(fread(buff, 1, BUFFSZ, fd) != BUFFSZ) read_err();
            if(fseek(fd, len - BUFFSZ, SEEK_CUR)) std_err();
            len = BUFFSZ;
        } else {
            if(fread(buff, 1, len, fd) != len) read_err();
        }
    }

    if(!msg[0]) return;
    buff[len] = 0;
    if(!dec) {
        printf(VISFMT"s %s\n", msg, buff);
    } else {
        p = cuteftp_pwd(buff, len, 0);
        if(p) printf(VISFMT"s %s\n", msg, p);

        p = cuteftp_pwd(buff, len, 1);
        if(p) printf(VISFMT"s %s\n", msg, p);
    }
}



u8 *cuteftp_pwd(u8 *pwd, u8 size, int algo) {
    static const u8 keyz[] = "kU6D4-1o_Q%3).k=+0976s 0sd0*Ojis4 )8duwijU";
    blf_ctx ctx;
    static u8   *tmp = NULL;
    int     i,
            j,
            len;
    u8      key[8],
            ivec[8],
            xivec[8],
            *p,
            *o;

    if(!algo) {
        tmp = realloc(tmp, size + 1);
        pwd = tmp;
        for(i = 0; i < size; i++) {
            pwd[i] ^= 0xc8;
        }
    } else {
        memcpy(key, "\x00\x01\x22\x00\x02\x00\x2B\x00", sizeof(key));
        for(i = 0; keyz[i]; i++) {
            key[i % sizeof(key)] ^= keyz[i];
        }

        len = strlen(pwd);
        tmp = realloc(tmp, ((len + 7) * 2) + 1);    // +7 to avoid padding problems
        o = tmp;
        for(p = pwd; *p; p++) {
            if(*p == '\\') {
                p++;
                if(!*p) {
                    *o++ = '\\';
                    break;
                }
                if(*p != '0') {
                    *o++ = '\\';
                    if(*p == '\\') continue;
                }
            }
            *o++ = *p;
        }
        while((o - tmp) % 8) *o++ = 0;

        pwd = tmp;
        size = o - tmp;
        memset(ivec, 0, 8);
        InitializeBlowfish(&ctx, key, sizeof(key));
        for(i = 0; i < size; i += 8) {
            memcpy(xivec, pwd + i, 8);
            Blowfish_decipher(&ctx, (void *)pwd + i, (void *)pwd + i + 4);
            for(j = 0; j < 8; j++) {
                pwd[i + j] ^= ivec[j];
                ivec[j] = xivec[j];
            }
        }
    }

    pwd[size] = 0;
    for(i = 0; i < size; i++) {
        if(!pwd[i]) break;
        if(pwd[i] == 0xc8) return(NULL);    // conflicts? rare
        if(pwd[i] < ' ')   return(NULL);
    }
    return(pwd);
}



void read_err(void) {
    printf("\nError: read error, this tool is experimental so probably a needed value has not been found\n");
    myexit();
}



void std_err(void) {
    perror("\nError");
    myexit();
}



void myexit(void) {
    printf("\n  press RETURN to exit\n");
    fgetc(stdin);
    exit(1);
}



